/*
** Copyright (c) 2003, Scott Lystig Fritchie.  All rights reserved.
** Some contributions Copyright (c) Chris Newcombe, 2006.  All rights reserved.
** See the file "../LICENSE" for license details.
*/

/* erl_driver_pipelib.h */

#include <sys/uio.h>
#include <erl_driver.h>
#include <sys/poll.h>


/* See comments in erl_driver_pipelib.c for why we have to copy this from Erlang VM source.
 * This is necessary in this headers (not just in erl_driver_pipelib.c) because
 * under R11B input_iovec_wrapper (below) needs to contain full Binary members rather
 * than just ErlDrvBinary (because the latter don't have a refc member any more).
 *   
 * MAINTENANCE ISSUE
 *
 * This structure macro is copied from R11B-1 otp_src_R11B-1/erts/emulator/beam/global.h.
 * This is the only way 
 */

/*
 * 'Binary' alignment:
 *   Address of orig_bytes[0] of a Binary should always be 8-byte aligned.
 * It is assumed that the flags, refc, and orig_size fields are 4 bytes on
 * 32-bits architectures and 8 bytes on 64-bits architectures.
 */

typedef struct binary {
    unsigned int flags;   /* cnewcom: was Uint */
    long refc;            /* cnewcom: is erts_refc_t in original, long in pipe-mode */
/* #ifdef ARCH_32 */
    unsigned int align__; /* cnewcom: was Uint32 */
/* #endif */
    /* Add fields BEFORE this, otherwise the drivers crash */
    long orig_size;
    char orig_bytes[1]; /* to be continued */
} Binary;

#define Binary2ErlDrvBinary(B) ((ErlDrvBinary *) (&((B)->orig_size)))
#define ErlDrvBinary2Binary(D) ((Binary *)        \
				(((char *) (D)) - \
				 ((char *) &(((Binary *) 0)->orig_size))))


/* driver_entry->outputv requires an ErlIOVec * holding the input.
 * read_packet_make_iovec() reads from the given fd and creates the ErlIOVec.
 *
 * The caller is responsible for calling edtk_driver_free_wrapper() on the result.
 *
 * (We cannot use a static buffer, as if 'non_block' entrypoints are called on
 * the Erlang side, and private_async_threadpool support is enabled,
 * we can get overlapping requests */

/* This is the header that goes at the start of the variable-sized buffer that
 * is returned by read_packet_make_input_iovec() */
typedef struct input_iovec_wrapper {
    ErlIOVec       ev;                /* this is what driver_entry->outputv actually needs */
    struct iovec   iov[2];
    ErlDrvBinary * binv[2];           /* binv[0] will be NULL,  binv[1] will be Binary2ErlDrvBinary(&input_data) (next member) */
    Binary         input_data;
                   /* the raw input data will be at input_data->orig_bytes */
} input_iovec_wrapper_t;

extern input_iovec_wrapper_t * read_packet_make_input_iovec(int fd);


extern unsigned int  num_fds_to_poll;
extern struct pollfd fds_to_poll[2];


/*Just for reference (from erl_driver.h)
 *
 * typedef struct erl_io_vec {
 *     int vsize;			length of vectors
 *     int size;			total size in bytes
 *     SysIOVec* iov;
 *     ErlDrvBinary** binv;
 * } ErlIOVec;
 *
 * typedef struct erl_drv_binary {
 *     long orig_size;       total length of binary
 *     long refc;            number of references to this binary
 *     char orig_bytes[1];   the data (char instead of byte!)
 * } ErlDrvBinary;
 *  
 */

