
#include <db.h>
#include "erl_interface.h"
#include "ei.h"
#include "berkeley_db_drv.h"

DB_ENV *my_db_env_create(u_int32_t, int *);
int my_db_env_close(DB_ENV *, u_int32_t);
int my_db_env_close_valmap_cleanup(DB_ENV *);
void my_msgcall(const DB_ENV *, const char *);
void my_errcall(const DB_ENV *, const char *, const char *);
int my_write_coredump(DB_ENV *dbenv, const char *filename);
int my_set_write_coredump_on_errcall(const DB_ENV *dbenv, int coredump_interval_secs, int regex_mode, const char * path_prefix, const char *regex);
void free_coredump_compiled_regex(descriptor_t *desc);
int my_env_set_errfile(DB_ENV *, char *, char *);
int my_db_set_errfile(DB *, char *, char *);
DB *my_db_create(DB_ENV *, u_int32_t, int *);
int my_db_env_open(DB_ENV *dbenv, const char *home_dir, u_int32_t flags, int mode);
int my_db_close_valmap_cleanup(DB *);
int my_db_close(DB *, u_int32_t);
DBC *my_cursor_create(DB *, DB_TXN *, u_int32_t, int *);
DBC *my_cursor_dup(DBC *, u_int32_t, int *);
int my_cursor_close_valmap_cleanup(DBC *);
DB_TXN *my_txn_begin(DB_ENV *, DB_TXN *, u_int32_t, int *);
int my_txn_abort_valmap_cleanup(DB_TXN *);
int my_db_verify(DB *, char *, char *, char *, u_int32_t);
int make_dbt_key(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount);
int make_dbt_data(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount);
void discard_descendents_of_prepared_txn(descriptor_t * desc, int parent_idx);
int my_make_list_unresolved_txns(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount);
int my_LOAD_unresolved_DistTxnId_and_valmap_record(descriptor_t *desc, int txn_idx, ErlDrvTermData *msg, int *members, int *msgcount);
void my_handle_sleepycat_event_notification(DB_ENV *, u_int32_t event, void *event_info);
void my_send_need_to_resolve_prepared_txn_notification(descriptor_t * desc, int txn_idx);
int check_is_safe_for_port_coordinator_to_close_db_handles(DB_ENV *dbenv);
int my_make_db_close_safe_if_there_are_unresolved_transactions(DB_ENV *dbenv, long *pout);
ETERM * make_rep_stat_external_term(descriptor_t *desc, callstate_t *c);
ETERM * make_lock_stat_external_term(descriptor_t *desc, callstate_t *c);
ETERM * make_mutex_stat_external_term(descriptor_t *desc, callstate_t *c);
ETERM * make_txn_stat_external_term(descriptor_t *desc, callstate_t *c);
ETERM * make_memp_stat_external_term(descriptor_t *desc, callstate_t *c);
ETERM * make_log_stat_external_term(descriptor_t *desc, callstate_t *c);
int my_db_stat(DB *db, DB_TXN *db_txn, DBTYPE * db_type_p, void **db_stat_p, u_int32_t flags);
ETERM * make_db_stat_external_term(descriptor_t *desc, callstate_t *c);
int make_db_compact_xreturn(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount);
void convert_bdb_return_code_to_driver_term(descriptor_t *desc, int error_code, ErlDrvTermData *msg, int *members, int *msgcount);
void convert_bdb_return_code_or_replication_error_to_driver_term(descriptor_t *desc, int error_code, ErlDrvTermData *msg, int *members, int *msgcount);
int make_dbtype(descriptor_t *desc, callstate_t *c,  ErlDrvTermData *msg, int *members, int *msgcount);
ETERM * make_log_filenames_external_term(descriptor_t *desc, callstate_t *c);
int my_rep_set_master_checkpoint_sleep(DB_ENV * dbenv, u_int32_t period_seconds); /* This is a temporary hack; it becomes part of rep_set_timeout in BDB v4.6 */
int my_env_interrupt_os_sleep(DB_ENV * dbenv);
int my_txn_commit(DB_TXN * db_txn, u_int32_t flags);
int my_txn_prepare(DB_TXN *db_txn, u_int8_t distributed_txn_guid[DB_XIDDATASIZE]);
int my_db_put(DB *db, DB_TXN *db_txn, DBT *key, DBT *data, u_int32_t flags);
int my_db_open(DB *db, DB_TXN *db_txn, const char *file, const char *database, DBTYPE type, u_int32_t flags, int mode);
int make_c_get_multiple_xreturn(descriptor_t *desc, callstate_t *c,  ErlDrvTermData *msg, int *members, int *msgcount);

