
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>

#ifdef	FREEBSD
#include <sys/time.h>
#include <net/bpf.h>
#include <sys/ioctl.h>
#endif	/* FREEBSD */

#ifdef	SOLARIS
#include <stropts.h>
#include <sys/filio.h>
#endif	/* SOLARIS */

#include <pcap.h>

#ifdef	DRIVER_USING_PTHREADS
#include <pthread.h>
static pthread_once_t _once = PTHREAD_ONCE_INIT;
#endif	/* DRIVER_USING_PTHREADS */

#include <erl_driver.h>
#include <libpcap_drv.h>
#include <erl_driver_tk.h>
#define	NEED_FUNC_PROTOS
#include <my-libpcap.h>

static int _initialized = 0;
static ErlDrvTermData am_pcap_if_t;
static ErlDrvTermData am_pcap_addr_t;
static ErlDrvTermData am_sockaddr;

void
null(void)
{
    printf("Hello, world!  This is null()\r\n");
}

/*
** For BSD systems, use ioctl(BIOCIMMEDIATE) to avoid waiting the full
** to_ms milliseconds before receiving a sporadic packet.
**
** QQQ TODO: conditional compilation for non-BSD systems.
*/

int
my_set_immediate(pcap_t *p, u_int immediate)
{
    return ioctl(pcap_fileno(p), BIOCIMMEDIATE, &immediate);
}

int
my_set_see_sent(pcap_t *p, int see_sent)
{
#if     defined(FREEBSD)
    u_int	zero = 0;

    zero = 0;
    if (ioctl(pcap_fileno(p), BIOCSSEESENT, &zero) < 0) {
	return -1;
    }
    zero = 5;
    if (ioctl(pcap_fileno(p), BIOCGSEESENT, &zero) < 0 || zero != 0) {
	return -1;
    }
    return 0;
#else  /* FREEBSD */
    /* QQQ XXX what should I do here?? */
    return 0;
#endif  /* FREEBSD */
}

/*
** QQQ I guess I didn't have to write this wrapper function if I:
**	1. Made it possible to stored a non-return-value in a valmap
**	2. Decided not to store bpf_program's in a valmap
** But I wanted to use a valmap, and I didn't bother extending 
** c_c_template.gsl to be able to store outarg thingies inside valmaps.
** The wrapper route takes a few more keystrokes, but perhaps it's the
** better thing to do??
*/

struct bpf_program *
my_pcap_compile(pcap_t *p, char *str, int optimize, bpf_u_int32 netmask)
{
    struct bpf_program	*fp;

    if ((fp = (struct bpf_program *) sys_alloc(sizeof(struct bpf_program))) == NULL) {
	return NULL;
    }
    if (pcap_compile(p, fp, str, optimize, netmask) < 0) {
	errno = EINVAL;		/* XXX Anything better?  {shrug} */
	return NULL;
    }
    return fp;
}

void my_pcap_freecode(struct bpf_program *fp)
{
    pcap_freecode(fp);
    sys_free(fp);
}

static void
init_my_atoms(void)
{
    am_pcap_if_t = driver_mk_atom("pcap_if_t");
    am_pcap_addr_t = driver_mk_atom("pcap_addr_t");
    am_sockaddr = driver_mk_atom("sockaddr");
    _initialized = 1;
}

int
make_findalldevs_list(struct descriptor *desc, callstate_t *c,
		      ErlDrvTermData *msg, int *members, int *msgcount)
{
    pcap_if_t	*ifp;
    int		if_s = 0;

    if (! _initialized) {
#ifdef	DRIVER_USING_PTHREADS
	pthread_once(&_once, init_my_atoms);
#else	/* DRIVER_USING_PTHREADS */
	init_my_atoms();
#endif	/* DRIVER_USING_PTHREADS */

    }

    for (ifp = c->o.alldevsp; ifp != NULL; ifp = ifp->next) {
	make_if_t(ifp, msg, msgcount);
	if_s++;
    }
    *msgcount = LOAD_NIL(msg, *msgcount);
    *members = if_s + 1;
    return if_s;
}

static int
make_if_t(pcap_if_t *ifp, ErlDrvTermData *msg, int *msgcount)
{
    int		mycount = 0;
    pcap_addr_t	*ap;
    char	*empty_string = "";
    int		aps = 0;

    *msgcount = LOAD_ATOM(msg, *msgcount, am_pcap_if_t);
    mycount++;

    *msgcount = LOAD_STRING(msg, *msgcount, ifp->name, strlen(ifp->name));
    mycount++;

    if (ifp->description == NULL) {
	*msgcount = LOAD_STRING(msg, *msgcount, empty_string, 0);
    } else {
	*msgcount = LOAD_STRING(msg, *msgcount, ifp->description, strlen(ifp->description));
    }
    mycount++;

    for (ap = ifp->addresses; ap != NULL; ap = ap->next) {
	if (ap->addr->sa_family == AF_INET ||
	    ap->addr->sa_family == AF_INET6 ||
#ifdef AF_LINK
	    ap->addr->sa_family == AF_LINK
#else  /* AF_LINK */
	    0
#endif /* AF_LINK */
						)  {
	    make_addr_t(ap, msg, msgcount);
	    aps++;
	} else {
	    edtk_debug("%s: skipping interface %s AF_ family %d", __FUNCTION__, ifp->name, ap->addr->sa_family);
	}
    }
    *msgcount = LOAD_NIL(msg, *msgcount);
    *msgcount = LOAD_LIST(msg, *msgcount, aps + 1); /* count the NIL! */
    mycount++;

    *msgcount = LOAD_INT(msg, *msgcount, ifp->flags);
    mycount++;

    *msgcount = LOAD_TUPLE(msg, *msgcount, mycount);

    return 1;
}

static int
make_addr_t(pcap_addr_t *ap, ErlDrvTermData *msg, int *msgcount)
{
    int		mycount = 0;

    *msgcount = LOAD_ATOM(msg, *msgcount, am_pcap_addr_t);
    mycount++;
    make_sockaddr(ap->addr, msg, msgcount);
    mycount++;
    make_sockaddr(ap->netmask, msg, msgcount);
    mycount++;
    make_sockaddr(ap->broadaddr, msg, msgcount);
    mycount++;
    make_sockaddr(ap->dstaddr, msg, msgcount);
    mycount++;

    *msgcount = LOAD_TUPLE(msg, *msgcount, mycount);

    return 1;
}

/*
** QQQ I don't know enough about the struct sockaddr here to be able to
** deal with these things sanely.  Here's what "ifconfig -a" says:
** 
** rl0: flags=8843<UP,BROADCAST,RUNNING,SIMPLEX,MULTICAST> mtu 1500
**         inet 10.1.1.1 netmask 0xffffff00 broadcast 10.1.1.255
**         inet6 fe80::230:1bff:fe07:8058%rl0 prefixlen 64 scopeid 0x1 
**         atalk 65280.223 range 0-65534 phase 2 broadcast 0.255
**         ether 00:30:1b:07:80:58 
**         media: Ethernet autoselect (100baseTX <full-duplex>)
**         status: active
** lo0: flags=8049<UP,LOOPBACK,RUNNING,MULTICAST> mtu 16384
**         inet6 ::1 prefixlen 128 
**         inet6 fe80::1%lo0 prefixlen 64 scopeid 0x4 
**         inet 127.0.0.1 netmask 0xff000000 
**         atalk 0.0 range 0-0 phase 2
** 
** ... and here's what I see if I just copy the sockaddr data as-is:
** 
** [{pcap_if_t,"rl0",
**             [],
**             [{pcap_addr_t,{sockaddr,18,		% Link layer
**                                     <<1,0,6,3,6,0,114,108,48,0,48,27,7,128,88,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0>>},
**                           {},
**                           {},
**                           {}},
**              {pcap_addr_t,{sockaddr,2,		% IPv4
**                                     <<0,0,10,1,1,1,0,0,0,0,0,0,0,0>>},
**                           {sockaddr,2,<<0,0,255,255,255>>},
**                           {sockaddr,2,
**                                     <<0,0,10,1,1,255,0,0,0,0,0,0,0,0>>},
**                           {}},
**              {pcap_addr_t,{sockaddr,28,		% IPv6
**                                     <<0,0,0,0,0,0,254,128,0,1,0,0,0,0,2,48,27,255,254,7,128,88,0,0,0,0>>},
**                           {sockaddr,28,
**                                     <<0,0,0,0,0,0,255,255,255,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0>>},
**                           {},
**                           {}}],
**             0},
**  {pcap_if_t,"lo0",
**             [],
**             [{pcap_addr_t,{sockaddr,18,		% Link layer
**                                     <<4,0,24,3,0,0,108,111,48,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0>>},
**                           {},
**                           {},
**                           {}},
**              {pcap_addr_t,{sockaddr,28,		% IPv6
**                                     <<0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0>>},
**                           {sockaddr,28,
**                                     <<0,0,0,0,0,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,0,0,0>>},
**                           {},
**                           {}},
**              {pcap_addr_t,{sockaddr,28,		% IPv6
**                                     <<0,0,0,0,0,0,254,128,0,4,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0>>},
**                           {sockaddr,28,
**                                     <<0,0,0,0,0,0,255,255,255,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0>>},
**                           {},
**                           {}},
**              {pcap_addr_t,{sockaddr,2,		% IPv4
**                                     <<0,0,127,0,0,1,0,0,0,0,0,0,0,0>>},
**                           {sockaddr,2,<<0,0,255>>},
**                           {},
**                           {}}],
**             1}]
** 
** I don't know how to deal with these sockaddrs.  I don't even know if
** the data in them is sane!  For example, the second IPv6 address on 
** "lo0" does not have a "4" in it, as my dumper says but "ifconfig"
** doesn't say.  {sigh}  I'll just leave this stuff as-is for now.
** If you want the address & mask of an interface, use pcap_lookupnet(),
** which I know works.  If you want to know the broadcast address or
** point-to-point peer's address, you're out of luck here; go write
** your own!  :-)
*/

static int
make_sockaddr(struct sockaddr *sp, ErlDrvTermData *msg, int *msgcount)
{
    int		mycount = 0;
    char	*bufp = NULL;
    char	*sa_data_start, *sa_data_end;
    int		sa_data_len;

    if (sp == NULL) {
	*msgcount = LOAD_TUPLE(msg, *msgcount, 0);
	return 1;
    }
    /* I hate this kind of offset arithmetic.... */
    sa_data_start = &sp->sa_data[0] + 2;
#if    defined(linux) || defined(SOLARIS)
    sa_data_end = (char *) sa_data_start + 4; /* QQQ What the heck goes here?? */
#else  /* linux || SOLARIS */
    sa_data_end = (char *) sa_data_start + sp->sa_len - 2;
#endif /* linux || SOLARIS */
    sa_data_len = sa_data_end - sa_data_start;
    if ((bufp = edtk_driver_alloc_wrapper(sa_data_len)) == NULL) {
	/*
	** QQQ We're just screwed here, given the stuff we've already
	** pushed onto the stack.  I think our options are to:
	** 1. abort, or 2. return prematurely.  If we do #2, then
	** driver_output_term() will puke, then the Erlang receiver
	** will hang forever waiting for a reply ... but at least the
	** rest of the VM will have a chance to try to continue.
	*/
	return 0;
    }
    *msgcount = LOAD_ATOM(msg, *msgcount, am_sockaddr);
    mycount++;
    memcpy(bufp, sa_data_start, sa_data_len);
    *msgcount = LOAD_INT(msg, *msgcount, sp->sa_family);
    mycount++;
    *msgcount = LOAD_BINARY(msg, *msgcount,
			    edtk_alloced_ptr2ErlDrvBinary(bufp),
			    0, sa_data_len);
    mycount++;
    *msgcount = LOAD_TUPLE(msg, *msgcount, mycount);

    return 1;
}
